from tulip_api.tulip_api import TulipAPI


class TulipTableLink:
    """
    An interface with Tulip Table Links.
    """

    def __init__(self, tulip_api: TulipAPI, link_id: str):
        self.tulip_api = tulip_api
        self.link_id = link_id

    def get_details(self):
        """
        GET `/tableLinks/{linkId}`

        Returns the metadata of a Tulip Table Link.
        """
        return self.tulip_api.make_request(
            self._construct_base_table_link_path(), "GET"
        )

    def link_records(self, left_record_id: str, right_record_id: str):
        """
        PUT `/tableLinks/{linkId}/link`

        Links the given records.
        """
        self.tulip_api.make_request_expect_nothing(
            self._construct_link_path(),
            "PUT",
            json={"leftRecord": left_record_id, "rightRecord": right_record_id},
        )

    def unlink_records(self, left_record_id: str, right_record_id: str):
        """
        PUT `/tableLinks/{linkId}/unlink`

        Unlinks the given records.
        """
        self.tulip_api.make_request_expect_nothing(
            self._construct_unlink_path(),
            "PUT",
            json={"leftRecord": left_record_id, "rightRecord": right_record_id},
        )

    def _construct_base_table_link_path(self):
        return f"tableLinks/{self.link_id}"

    def _construct_link_path(self):
        return f"{self._construct_base_table_link_path()}/link"

    def _construct_unlink_path(self):
        return f"{self._construct_base_table_link_path()}/unlink"



from typing import Dict, List

from tulip_api.exceptions import (
    TulipAPICachedTableDuplicateIDFound,
    TulipApiCachedTableRecordNotFound,
)
from tulip_api.tulip_api import TulipAPI
from tulip_api.tulip_table import TulipTable


class CachedTulipTable:
    """
    Pulls a given table/filter into memory. Reference the List `.records` or use the `get_record` method to get a specific record by it's ID.

    Use with caution and only with small tables.
    """

    def __init__(self, tulip_api: TulipAPI, table_id: str, filters: List = []):
        self.tulip_api = tulip_api
        self.table_id = table_id
        self.filters = filters
        self.tulip_table = TulipTable(self.tulip_api, self.table_id)
        self.update_data()

    def _fetch_data(self) -> List:
        return list(self.tulip_table.stream_records(filters=self.filters))

    def update_data(self):
        self.records = self._fetch_data()

    def get_record(self, record_id: str) -> Dict:
        records = list(filter(lambda record: record["id"] == record_id, self.records))
        records_length = len(records)
        if records_length == 1:
            return records[0]
        elif records_length == 0:
            raise TulipApiCachedTableRecordNotFound(record_id)
        else:
            raise TulipAPICachedTableDuplicateIDFound(record_id)



import asyncio
from typing import Any, AsyncGenerator, Iterable, List, Union
from uuid import uuid4

from tulip_api.asyncio import TulipAPI
from tulip_api.exceptions import (
    TulipAPIInvalidChunkSize,
    TulipApiTableRecordCreateMustIncludeID,
)


class TulipTable:
    """
    An interface with a Tulip Table.
    """

    def __init__(self, tulip_api: TulipAPI, table_id: str):
        self.tulip_api = tulip_api
        self.table_id = table_id

    async def get_details(self):
        """
        GET `/tables/{tableId}`

        Gets details about a Tulip Table's metadata and schema.
        """
        return await self.tulip_api.make_request(self._construct_base_path(), "GET")

    async def update_table(
        self,
        label: Union[str, None] = None,
        description: Union[str, None] = None,
        deleted: bool = False,
        new_columns: List = [],
        hide_columns: List[str] = [],
        unhide_columns: List[str] = [],
    ):
        """
        PUT `/tables/{tableId}`

        Update the tables metadata or schema.
        This method pulls the table's current metadata and schema before updating with any values passed in.

        new_columns: A list of columns to add to the table schema

        hide_columns: A list of columnId's to mark as hidden.

        unhide_columns: A list of columnId's to unmark as hidden.
        """
        table = await self.get_details()
        if label:
            table["label"] = label
        if description:
            table["description"] = description

        for column in table["columns"]:
            if column["name"] in hide_columns:
                column["hidden"] = True
            if column["name"] in unhide_columns:
                column["hidden"] = False

        table["columns"] += new_columns

        return await self.tulip_api.make_request(
            self._construct_base_path(),
            "PUT",
            json={
                "label": table["label"],
                "description": table["description"],
                "deleted": deleted,
                "columns": table["columns"],
            },
        )

    async def get_records(
        self,
        limit: int = 100,
        offset: int = 0,
        filters: List = [],
        sort_by: str = "_updatedAt",
        sort_asc: bool = False,
        filter_aggregator: str = "all",
    ):
        """
        GET `/tables/{tableId}/records`

        Get a list of table records.

        `filters`: A list of filters
        ```
        {
          "field": "{field}",
          "arg": "{arg}",
          "functionType": "{functionType}",
        }
        ```
        """
        params = {
            "limit": limit,
            "offset": offset,
            "sortBy": sort_by,
            "filterAggregator": filter_aggregator,
            "sortDir": "asc" if sort_asc else "desc",
        }
        for index, filter in enumerate(filters):
            for key, value in filter.items():
                params[f"filters.{index}.{key}"] = value
        return await self.tulip_api.make_request(
            self._construct_records_path(), "GET", params=params
        )

    async def stream_records(
        self,
        filters: List = [],
        sort_by: str = "_updatedAt",
        sort_asc: bool = False,
        filter_aggregator: str = "all",
        chunk_size: int = 100,
        limit: Union[int, None] = None,
    ) -> AsyncGenerator[dict, None]:
        """
        Returns a Generator that will pull all (or up to a limit) records from a Tulip Table.

        `chunk_size`: Must be between 1 and 100
        """
        if chunk_size < 1 or chunk_size > 100:
            raise TulipAPIInvalidChunkSize(chunk_size)
        offset = 0
        index = 0
        while True:
            records = await self.get_records(
                limit=chunk_size,
                offset=offset,
                filters=filters,
                sort_by=sort_by,
                sort_asc=sort_asc,
                filter_aggregator=filter_aggregator,
            )
            if len(records) == 0:
                break

            for record in self._stream_records_helper(records, limit, index):
                index += 1
                yield record

            offset += chunk_size

            if index < offset:
                break

    @staticmethod
    def _stream_records_helper(
        records: List, limit: Union[int, None], index
    ) -> Iterable:
        if limit is None:
            for record in records:
                yield record
            return

        for record in records:
            if index > limit:
                return
            index += 1
            yield record

    async def get_record(self, record_id: str):
        """
        GET `/tables/{tableId}/records/{recordId}`

        Returns a record from a Tulip Table given it's ID
        """
        return await self.tulip_api.make_request(
            self._construct_record_path(record_id), "GET"
        )

    async def get_record_by_sort_filter(
        self,
        filters: List = [],
        sort_by: str = "_updatedAt",
        sort_asc: bool = False,
    ) -> Union[dict, None]:
        """
        GET `/tables/{tableId}/records`

        Returns the first record from a Tulip Table `get_records` request.
        """
        records = await self.get_records(
            limit=1, filters=filters, sort_by=sort_by, sort_asc=sort_asc
        )
        if len(records) != 1:
            return None
        return records[0]

    async def create_record(self, record: dict, create_random_id=False):
        """
        POST `/tables/{tableId}/records`

        Create a new record in a Tulip Table.
        Either `record` must contain `id` or the `create_random_id` flag must be set to `True`.

        Returns the created record.
        """
        if create_random_id:
            record["id"] = uuid4().hex
        if "id" not in record:
            raise TulipApiTableRecordCreateMustIncludeID()
        return await self.tulip_api.make_request(
            self._construct_records_path(), "POST", json=record
        )

    async def create_records(
        self, records: Iterable[dict], create_random_id=False, warn_on_failure=False
    ) -> int:
        """
        Iterates over a list of records and creates them. Calling `create_record`

        Returns the # of successfully created records.

        `warn_on_failure`: set to True if you want to continue with creating the rest of the records
        , despite a malformed request.

        """
        created_records = 0
        failed_records = 0
        futures: List[asyncio.Task] = []
        for record in records:
            futures.append(
                asyncio.create_task(
                    self.create_record(record, create_random_id=create_random_id)
                )
            )

        for future in asyncio.as_completed(futures):
            try:
                await future
                created_records += 1
            except Exception as e:
                failed_records += 1
                print(f"There was an issue creating a record\n{e}")
                if not warn_on_failure:
                    raise e

        if warn_on_failure and failed_records > 0:
            print(f"Failed to create {failed_records} records.")

        return created_records

    async def update_record(self, record_id: str, record: dict = {}):
        """
        PUT `/tables/{tableId}/records/{recordId}`

        Updates a Tulip Table record. Only modifies the passed in columns.
        """
        return await self.tulip_api.make_request(
            self._construct_record_path(record_id), "PUT", json=record
        )

    async def delete_record(self, record_id: str):
        """
        DELETE `/tables/{tableId}/records/{recordId}`

        Deletes a Tulip Table record.

        Returns the deleted record.
        """
        return await self.tulip_api.make_request(
            self._construct_record_path(record_id), "DELETE"
        )

    async def delete_records(self, allow_records_in_use=False):
        """
        DELETE `/tables/{tableId}/records`

        Deletes all records from a Tulip Table.
        """
        # lowercase boolean because the api doesn't like capitalized booleans.
        return await self.tulip_api.make_request_expect_nothing(
            self._construct_records_path(),
            "DELETE",
            params={"allowRecordsInUse": "true" if allow_records_in_use else "false"},
        )

    async def increment_record_column(self, record_id: str, column_id: str, value: int):
        """
        PATCH `/tables/{tableId}/records/{recordId}/increment`

        Increments a given record/column's value by the given value.
        """
        return await self.tulip_api.make_request(
            self._construct_increment_record_path(record_id),
            "PATCH",
            json={"fieldName": column_id, "value": value},
        )

    @staticmethod
    def coalesce(value: Any, default: Any = 0):
        """
        A helper method to wrap possibly null table column values with a default value.

        EX:
        ```
        TulipTable.coalesce(record['a_column']) + 1
        ```
        """
        if not value:
            return default
        return value

    def _construct_base_path(self):
        return f"tables/{self.table_id}"

    def _construct_records_path(self):
        return f"{self._construct_base_path()}/records"

    def _construct_record_path(self, record_id: str):
        return f"{self._construct_records_path()}/{record_id}"

    def _construct_increment_record_path(self, record_id: str):
        return f"{self._construct_record_path(record_id)}/increment"



import aiohttp
import requests


class BaseTulipAPIException(Exception):
    """All custom exceptions from within this package inherit from this exception"""

    pass


class TulipAPINoCredentialsFound(BaseTulipAPIException):
    """No credential provider was able to provide credentials."""

    pass


class TulipAPIAuthorizationError(BaseTulipAPIException):
    """The given credentials were unable to authenticate with the tulip instance."""

    def __init__(self, response: requests.Response):
        self.message = (
            f"The {response.request.method} request to {response.url} was not able to authenticate using the given credentials.\n"
            f"Response status code: {response.status_code}."
        )
        super().__init__(self.message)


class TulipAPIAsyncAuthorizationError(BaseTulipAPIException):
    """The given credentials were unable to authenticate with the tulip instance."""

    def __init__(self, response: aiohttp.ClientResponse):
        self.message = (
            f"The {response.method} request to {response.url} was not able to authenticate using the given credentials.\n"
            f"Response status code: {response.status}."
        )
        super().__init__(self.message)


class TulipAPIConnectionError(BaseTulipAPIException):
    """Unable to connect to the given tulip instance."""

    pass


class TulipAPIMalformedRequestError(BaseTulipAPIException):
    """The request was malformed"""

    def __init__(self, response: requests.Response):
        self.message = (
            f"The {response.request.method} request to {response.url} was malformed.\n"
            f"Response status code: {response.status_code}."
        )
        super().__init__(self.message)


class TulipAPIAsyncMalformedRequestError(BaseTulipAPIException):
    """The request was malformed"""

    def __init__(self, response: aiohttp.ClientResponse):
        self.message = (
            f"The {response.method} request to {response.url} was malformed.\n"
            f"Response status code: {response.status}."
        )
        super().__init__(self.message)


class TulipAPINotFoundError(BaseTulipAPIException):
    """The requested resource was not found"""

    def __init__(self, response: requests.Response):
        self.message = (
            f"The {response.request.method} request to {response.url} did not find the requested resource.\n"
            f"Response status code: {response.status_code}."
        )
        super().__init__(self.message)


class TulipAPIAsyncNotFoundError(BaseTulipAPIException):
    """The requested resource was not found"""

    def __init__(self, response: aiohttp.ClientResponse):
        self.message = (
            f"The {response.method} request to {response.url} did not find the requested resource.\n"
            f"Response status code: {response.status}."
        )
        super().__init__(self.message)


class TulipAPIInternalError(BaseTulipAPIException):
    """The requested resource was not found"""

    def __init__(self, response: requests.Response):
        self.message = (
            f"The {response.request.method} request to {response.url} resulted in an internal error.\n"
            f"Response status code: {response.status_code}.\n"
            "Request Body:\n"
            f"{response.request.body}"
        )
        super().__init__(self.message)


class TulipAPIAsyncInternalError(BaseTulipAPIException):
    """The requested resource was not found"""

    def __init__(self, response: aiohttp.ClientResponse):
        self.message = (
            f"The {response.method} request to {response.url} resulted in an internal error.\n"
            f"Response status code: {response.status}."
        )
        super().__init__(self.message)


class TulipAPIUnknownResponse(BaseTulipAPIException):
    """The requested resource was not found"""

    def __init__(self, response: requests.Response):
        self.message = (
            f"The {response.request.method} request to {response.url} resulted in an unknown response.\n"
            f"Response status code: {response.status_code}."
        )
        super().__init__(self.message)


class TulipAPIAsyncUnknownResponse(BaseTulipAPIException):
    """The requested resource was not found"""

    def __init__(self, response: aiohttp.ClientResponse):
        self.message = (
            f"The {response.method} request to {response.url} resulted in an unknown response.\n"
            f"Response status code: {response.status}."
        )
        super().__init__(self.message)


class TulipAPIInvalidChunkSize(BaseTulipAPIException):
    """Tulip Table record limit must be between 1 and 100"""

    def __init__(self, chunk_size: int):
        self.message = f"Chunk Size must be between 1 and 100. {chunk_size} is invalid."


class TulipAPICachedTableDuplicateIDFound(BaseTulipAPIException):
    """Multiple records with the same id were found in a cached table"""

    def __init__(self, record_id: str):
        self.message = "Multiple records were found with the id {record_id}"
        super().__init__(self.message)


class TulipApiCachedTableRecordNotFound(BaseTulipAPIException):
    """Record was not found in a cached tulip table"""

    def __init__(self, record_id: str):
        self.message = "No record found with the id {record_id}"
        super().__init__(self.message)


class TulipApiTableRecordCreateMustIncludeID(BaseTulipAPIException):
    def __init__(self):
        self.message = "Table Record creates must include an `id` key in the record, or the `create_random_id` flag must be set to True."



from typing import Any, Dict

from tulip_api.asyncio.tulip_api import TulipAPI


class TulipMachine:
    """
    An interface with the machine event reporting api.
    """

    attributes_report_path: str = "attributes/report"

    def __init__(self, tulip_api: TulipAPI, machine_id: str):
        self.tulip_api = tulip_api
        self.machine_id = machine_id

    async def send_event(self, attributes: Dict[str, Any]):
        """
        POST `/attributes/report`

        `attributes`: A dict with attributeId: value pairs.
        ```
        {
            "attributeIdA": "valueA",
            "attributeIdB": "valueB",
        }
        ```
        """
        await self.tulip_api.make_request_expect_nothing(
            TulipMachine.attributes_report_path,
            "POST",
            json=self._construct_attributes(attributes),
        )

    def _construct_attributes(self, attributes: Dict[str, Any]):
        return {
            "attributes": [
                {"machineId": self.machine_id, "attributeId": key, "value": value}
                for key, value in attributes.items()
            ]
        }



from csv import DictReader
from typing import Any, Dict, Iterable, TextIO, Union

from dateutil import parser

from tulip_api.asyncio.tulip_table import TulipTable


class TulipTableCSVUploader:
    tulip_table: TulipTable

    def __init__(self, tulip_table: TulipTable, csv_file: Union[str, TextIO]):
        self.tulip_table = tulip_table
        self.csv_file = csv_file

    async def execute(self, create_random_id=False, warn_on_failure=False) -> int:
        if isinstance(self.csv_file, str):
            with open(self.csv_file, "r") as csv_file:
                return await self._upload_records(
                    csv_file,
                    create_random_id=create_random_id,
                    warn_on_failure=warn_on_failure,
                )

        return await self._upload_records(
            self.csv_file,
            create_random_id=create_random_id,
            warn_on_failure=warn_on_failure,
        )

    async def _upload_records(
        self, file: TextIO, create_random_id=False, warn_on_failure=False
    ):
        table_columns = (await self.tulip_table.get_details())["columns"]
        column_types = {
            column["name"]: column["dataType"]["type"] for column in table_columns
        }
        reader = DictReader(file)
        TulipTableCSVUploader._validate_csv_columns(reader.fieldnames, column_types)

        return await self.tulip_table.create_records(
            TulipTableCSVUploader._yield_coerced_records(reader, column_types),
            warn_on_failure=warn_on_failure,
            create_random_id=create_random_id,
        )

    @staticmethod
    def _coerce_type(value: Any, type: str):
        if type == "string":
            return str(value)
        if type == "integer":
            if isinstance(value, str):
                return int(float(value))
            return int(value)
        if type == "float":
            return float(value)
        if type == "boolean":
            return bool(value)
        if type == "timestamp":
            return parser.parse(value, ignoretz=True).strftime("%Y-%m-%dT%H:%M:%SZ")

        raise Exception("Unsupported datatype: {type}. Value: {value}")

    @staticmethod
    def _coerce_record_types(record: Dict[str, Any], column_types: Dict[str, str]):
        new_record = {}
        for column_id, value in record.items():
            if column_id not in column_types:
                raise Exception(
                    f"Column {column_id} found in record, but not in table."
                )
            new_record[column_id] = TulipTableCSVUploader._coerce_type(
                value, column_types[column_id]
            )
        return new_record

    @staticmethod
    def _yield_coerced_records(records: Iterable, column_types: Dict[str, str]):
        for record in records:
            yield TulipTableCSVUploader._coerce_record_types(record, column_types)

    @staticmethod
    def _validate_csv_columns(csv_fieldnames, table_columns):
        for csv_fieldname in csv_fieldnames:
            if not csv_fieldname in table_columns:
                raise Exception(
                    f"Column {csv_fieldname} is not found in the Tulip Table columns."
                )



import os

from setuptools import setup


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="community-tulip-api",
    author="Henry Jones",
    author_email="henryivesjones@gmail.com",
    description="A community wrapper for the Tulip API",
    url="https://github.com/henryivesjones/community-tulip-api",
    packages=["tulip_api", "tulip_api.asyncio"],
    package_dir={"tulip_api": "tulip_api", "tulip_api.asyncio": "tulip_api/asyncio"},
    install_requires=["requests", "aiohttp", "python-dateutil"],
    long_description=read("README.md"),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Utilities",
        "LICENSE :: OSI APPROVED :: GNU AFFERO GENERAL PUBLIC LICENSE V3",
    ],
)



from tulip_api.cached_tulip_table import CachedTulipTable
from tulip_api.tulip_api import TulipAPI
from tulip_api.tulip_machine import TulipMachine
from tulip_api.tulip_table import TulipTable
from tulip_api.tulip_table_csv_upload import TulipTableCSVUploader
from tulip_api.tulip_table_link import TulipTableLink



import os
from base64 import b64encode
from typing import Any, Union

import aiohttp

from tulip_api.exceptions import (
    TulipAPIAsyncAuthorizationError,
    TulipAPIAsyncInternalError,
    TulipAPIAsyncMalformedRequestError,
    TulipAPIAsyncNotFoundError,
    TulipAPIAsyncUnknownResponse,
    TulipAPINoCredentialsFound,
)
from tulip_api.tulip_api import TulipAPIResponseCodes


class TulipAPI:
    """
    Asynio enabled

    Wraps the `self.session.` function with authentication, response processing, and base url construction.
    """

    def __init__(
        self,
        tulip_url: str,
        concurrency: int = 40,
        api_key: Union[str, None] = None,
        api_key_secret: Union[str, None] = None,
        auth: Union[str, None] = None,
        use_full_url: bool = False,
    ):
        """
        use_full_url: if set to true, the tulip_url must include `http://` or `https://` as well as the fqdn. For example `https://abc.tulip.co`
        """
        self.host = self._construct_base_url(tulip_url, use_full_url)

        self.auth = TulipAPI._provide_api_credentials(
            api_key=api_key, api_key_secret=api_key_secret, auth=auth
        )

        self.headers = self._construct_headers()
        self.concurrency = concurrency

    def __enter__(self):
        self.connector = aiohttp.TCPConnector(limit=self.concurrency)
        return self

    def __exit__(self, _, __, ___):
        self.connector.close()

    async def make_request(
        self,
        path: str,
        method: str,
        params: Union[dict, None] = None,
        json: Any = None,
    ):
        """
        Makes a request against the Tulip API. Parses and returns JSON returned from the Tulip API.
        """
        async with aiohttp.request(
            method,
            self._construct_url(path),
            params=params,
            json=json,
            headers=self.headers,
            connector=self.connector,
        ) as response:
            return await self._handle_api_response(response).json()

    async def make_request_expect_nothing(
        self,
        path: str,
        method: str,
        params: Union[dict, None] = None,
        json: Any = None,
    ):
        """
        Makes a request against the Tulip API. Returns nothing.
        """
        async with aiohttp.request(
            method,
            self._construct_url(path),
            params=params,
            json=json,
            headers=self.headers,
        ) as response:
            self._handle_api_response(response)

    @staticmethod
    def _provide_api_credentials(
        api_key: Union[str, None] = None,
        api_key_secret: Union[str, None] = None,
        auth: Union[str, None] = None,
    ):
        if auth:
            return auth
        if api_key and api_key_secret:
            return b64encode(f"{api_key}:{api_key_secret}".encode("utf-8")).decode(
                "utf-8"
            )
        if "TULIP_AUTH" in os.environ:
            return os.environ["TULIP_AUTH"]
        raise TulipAPINoCredentialsFound()

    @staticmethod
    def _construct_base_url(host, use_full_url):
        if use_full_url:
            return f"{host}/api/v3"
        cleaned_host = host.replace("http://", "").replace("https://", "")
        return f"https://{cleaned_host}/api/v3/"

    def _construct_url(self, path):
        return f"{self.host}{path}"

    def _construct_headers(self):
        return {"Authorization": f"Basic {self.auth}"}

    def _handle_api_response(self, response: aiohttp.ClientResponse):
        if response.status in TulipAPIResponseCodes.SUCCESS_CODES:
            return response
        if response.status in TulipAPIResponseCodes.MALFORMED_CODES:
            raise TulipAPIAsyncMalformedRequestError(response)
        if response.status in TulipAPIResponseCodes.NOT_FOUND_CODES:
            raise TulipAPIAsyncNotFoundError(response)
        if response.status in TulipAPIResponseCodes.UNAUTHENTICATED_CODES:
            raise TulipAPIAsyncAuthorizationError(response)
        if response.status in TulipAPIResponseCodes.UNEXCPECTED_ERROR_CODES:
            raise TulipAPIAsyncInternalError(response)
        raise TulipAPIAsyncUnknownResponse(response)
